import numpy as np

import sold.allocate


def test_first_price_allocation():
    bids = (5, 3, 2, 1, 8)
    allocation = sold.allocate.first_price(bids=bids)
    assert np.array_equal(allocation, np.array([0, 0, 0, 0, 1]))


def test_second_price_allocation_with_tie():
    """
    With a tie, the allocation is given to the first
    """
    bids = (5, 3, 2, 1, 8, 2, 8)
    allocation = sold.allocate.first_price(bids=bids)
    assert np.array_equal(allocation, np.array([0, 0, 0, 0, 1, 0, 0]))



import numpy as np


def auction(
    valuation_distributions,
    bidding_functions,
    allocation_rule,
    payment_rule,
    seed=None,
):
    """
    Runs a simulated auction

    Parameters
    ----------
    valuation_distributions: iterable
        A collection of N scipy.stats distributions (or any instance with a `rvs` method)
    bidding_functions: iterable
        A collection of N bidding functions that maps a value to a bid.
        All bidding functions must have a `value` keyword in its signature.
    allocation_rule: func
        A function that takes the collection of bids and returns an N simplex of allocations
        This function must have a `bids` keyword in its signature.
    payment_rule: func
        A function that takes the allocations and returns the payment made by each player
        This function must have `bids` keyword in its signature.
    seed: int
        A seed for the random number generation

    Returns
    -------
    allocation: an element of the simplex on N
    payments: the payments made by each individual
    valuations: the sampled valuations of each individual
    """
    assert len(valuation_distributions) == len(
        bidding_functions
    ), "Incorrect number of valuation distribution and bidding functions"

    np.random.seed(seed)
    valuations = [distribution.rvs() for distribution in valuation_distributions]
    bids = [
        function(value=value) for function, value in zip(bidding_functions, valuations)
    ]
    allocation = allocation_rule(bids=bids)
    payments = payment_rule(bids=bids)
    return np.array(allocation), np.array(payments), np.array(valuations)



import scipy.stats
import numpy as np

import sold
import sold.allocate
import sold.bid
import sold.pay


def test_auction_with_seed_0():
    N = 2
    seed = 0
    valuation_distributions = [scipy.stats.uniform() for _ in range(N)]
    bidding_functions = [sold.bid.true_value for _ in range(N)]
    allocation, payments, valuations = sold.auction(
        valuation_distributions=valuation_distributions,
        bidding_functions=bidding_functions,
        allocation_rule=sold.allocate.first_price,
        payment_rule=sold.pay.first_price,
        seed=seed,
    )
    assert np.array_equal(allocation, [0, 1])
    assert np.allclose(payments, [0, 0.71518937])
    assert np.allclose(valuations, [0.5488135, 0.71518937])


def test_auction_with_seed_2():
    N = 2
    seed = 2
    valuation_distributions = [scipy.stats.uniform() for _ in range(N)]
    bidding_functions = [sold.bid.true_value for _ in range(N)]
    allocation, payments, valuations = sold.auction(
        valuation_distributions=valuation_distributions,
        bidding_functions=bidding_functions,
        allocation_rule=sold.allocate.first_price,
        payment_rule=sold.pay.first_price,
        seed=seed,
    )
    assert np.array_equal(allocation, [1, 0])
    assert np.allclose(payments, [0.4359949, 0])
    assert np.allclose(valuations, [0.4359949, 0.02592623])


def test_auction_works_with_seed_function():
    """
    Create an auction with 3 players, two use a true value and the third use the
    theoretic weakly dominated strategy of 2/3 value.

    In expectation this third player should do better.
    """
    N = 3
    repetitions = 100
    valuation_distributions = [scipy.stats.uniform() for _ in range(N)]
    bidding_functions = [sold.bid.true_value for _ in range(N - 1)] + [
        sold.bid.create_shaded_bid_map(2 / 3)
    ]
    utilities = []
    for seed in range(repetitions):
        allocation, payments, valuations = sold.auction(
            valuation_distributions=valuation_distributions,
            bidding_functions=bidding_functions,
            allocation_rule=sold.allocate.first_price,
            payment_rule=sold.pay.first_price,
            seed=seed,
        )
        utilities.append(valuations - allocation * payments)
    mean_utilities = np.mean(utilities, axis=0)
    assert np.allclose(mean_utilities, [0.17535816, 0.18842919, 0.47175621])



import numpy as np


def first_price(bids):
    """
    Returns an allocation giving the item to the highest bid.

    Note that in the case of ties it returns the first highest bid.

    Parameters
    ----------
    bids: iterable
        The collection of bids

    Returns
    -------
    np array
    """
    bids = np.array(bids)
    first_price = np.max(bids)
    winner_index = np.where(bids == first_price)[0][0]
    allocation = np.zeros_like(bids)
    allocation[winner_index] = 1
    return allocation



import numpy as np

import sold.pay


def test_first_price_pay_function():
    bids = (5, 3, 2, 1, 8)
    payments = sold.pay.first_price(bids)
    assert np.array_equal(payments, np.array((0, 0, 0, 0, 8)))


def test_first_price_pay_function_with_ties():
    bids = (5, 3, 2, 1, 8, 2, 8)
    payments = sold.pay.first_price(bids)
    assert np.array_equal(payments, np.array((0, 0, 0, 0, 8, 0, 0)))


def test_second_price_pay_function():
    bids = (5, 3, 2, 1, 8)
    payments = sold.pay.second_price(bids)
    assert np.array_equal(payments, np.array((0, 0, 0, 0, 5)))



import numpy as np


def first_price(bids):
    """
    Returns the payments for a first price payment auction

    Parameters
    ----------
    bids: iterable
        The collection of bids

    Returns
    -------
    np array
    """
    bids = np.array(bids)
    first_price = np.max(bids)
    winner_index = np.where(bids == first_price)[0][0]
    payments = np.zeros_like(bids)
    payments[winner_index] = first_price
    return payments


def second_price(bids):
    """
    Returns the payments for a second price payment auction

    Parameters
    ----------
    bids: iterable
        The collection of bids

    Returns
    -------
    np array
    """
    bids = np.array(bids)
    first_price = np.max(bids)
    second_price = np.max([bid for bid in bids if bid != first_price])
    winner_index = np.where(bids == first_price)[0][0]
    payments = np.zeros_like(bids)
    payments[winner_index] = second_price
    return payments



import sold.bid


def test_true_bid():
    """
    A test for the true bid function
    """
    value = 50
    bid = sold.bid.true_value(value)
    assert value == bid


def test_create_shaded_bid():
    """
    A text to check the generation of a shaded bid function
    """
    value = 1
    shade = 0.8
    bid_map = sold.bid.create_shaded_bid_map(shade=shade)
    bid = bid_map(value=value)
    assert bid == shade * value



"""
Test that the version number exists and is valid.
"""

import packaging.version

import sold


def test_version_is_str():
    assert type(sold.__version__) is str


def test_version_is_valid():
    assert type(packaging.version.parse(sold.__version__)) is packaging.version.Version



def true_value(value):
    """
    Returns the value as the bid

    Parameters
    ----------
    value: float
        The value of the good being auctioned.

    Returns
    -------
    float
    """
    return value


def create_shaded_bid_map(shade):
    """
    Returns a function that maps a value to shade * value

    Parameters
    ----------
    shade: float
        A scaling parameter, usually between 0 and 1.
    Returns
    -------
    func
    """

    def bid_map(value):
        return shade * value

    return bid_map
